/////////////////////////////////////////////////////////
// Conditional values
/////////////////////////////////////////////////////////
#define COND_EQ  0	// Z set
#define COND_NE  1	// Z clear
#define COND_CS  2	// C set
#define COND_CC  3	// C clear
#define COND_MI  4	// N set
#define COND_PL  5	// N clear
#define COND_VS  6	// V set
#define COND_VC  7	// V clear
#define COND_HI  8	// C set and Z clear
#define COND_LS  9	// C clear or Z set
#define COND_GE 10	// N equals V
#define COND_LT 11	// N not equal to V
#define COND_GT 12	// Z clear AND (N equals V)
#define COND_LE 13	// Z set OR (N not equal to V)
#define COND_AL 14	// Always
#define COND_NV 15	// Never execute

/////////////////////////////////////////////////////////
// Data-operation constant building functions
/////////////////////////////////////////////////////////
#define SHIFT_LSL 0
#define SHIFT_LSR 1
#define SHIFT_ASR 2
#define SHIFT_ROR 3

// Generated value is (rot rotated right by (immed * 2))
#define armBuildShiftImm(rot, immed) \
	((rot) << 8) | (immed)
// RORX is generated with shift value of zero and SHIFT_ROR
#define armBuildShiftRegImm(shift, type, rm) \
	((shift) << 7) | ((type) << 5) | (rm)
#define armBuildShiftRegReg(rs, type, rm) \
	((rs) << 8) | ((type) << 5) | (1 << 4) | (rm)

#define OP_AND 0
#define OP_EOR 1
#define OP_SUB 2
#define OP_RSB 3
#define OP_ADD 4
#define OP_ADC 5
#define OP_SBC 6
#define OP_RSC 7
#define OP_TST 8
#define OP_TEQ 9
#define OP_CMP 10
#define OP_CMN 11
#define OP_ORR 12
#define OP_MOV 13
#define OP_BIC 14
#define OP_MVN 15

/////////////////////////////////////////////////////////
// OP is OP_***
// imm is 1 for immediate mode, 0 for register
// flags is 1 for update flags, 0 for don't
/////////////////////////////////////////////////////////
#define armBuildDataOpcode(op, imm, flags) \
	((imm) << 5) | ((op) << 1) | (flags)

#define armEmitDataOp(cond, opcode, rn, rd, shift) \
	((cond) << 28) | ((opcode) << 20) | ((rn) << 16) | ((rd) << 12) | (shift)

/////////////////////////////////////////////////////////
// Data Load/Store generation commands
/////////////////////////////////////////////////////////

// Immediate offset/index
// Register offset/index (same as scaled, just no scaling)
// Scaled register offset/index

#define armEmitLoadStoreWordAddressingImmediate(imm) (imm)
#define armEmitLoadStoreWordAddressingRegister(rm) (rm)
#define armEmitLoadStoreWordAddressingScaledRegister(amount, type, rm) \
	((amount) << 7) | ((type) << 5) | (rm)

// LDR
// I register offset or no I immediate offset
// B byte or no B word
// P post-indexed or no P pre-indexed
// U (offset add) or no U (offset sub)
// W (writeback if P == 1), or no W (no writeback)
// L (load) or no L (store)

#define armEmitLoadStoreWord(cond, i, p, u, b, w, l, rn, rd, addressing) \
	((cond) << 28) | (0x1 << 26) | ((i) << 25) | ((p) << 24) | ((u) << 23) | ((b) << 22) | \
	((w) << 21) | ((l) << 20) | ((rn) << 16) | ((rd) << 12) | (addressing)

// Halfword type

#define armEmitLoadStoreHalfwordAddressingImmediate(imm) (imm)
#define armEmitLoadStoreHalfwordAddressingRegister(rm) (rm)

#define armEmitHalfwordType 0xB
#define armEmitSignedByteType 0xD
#define armEmitSignedHalfwordType 0xF

#define armEmitLoadStoreHalfword(cond, p, u, i, w, l, rn, rd, addr, type) \
	((cond) << 28) | ((p) << 24) | ((u) << 23) | ((i) << 22) | ((w) << 21) | ((l) << 20) | \
	((rn) << 16) | ((rd) << 12) | (((addr) & 0xF0) << 4) | ((type) << 4) | ((addr) & 0xF)

/////////////////////////////////////////////////////////
// Branch commands
/////////////////////////////////////////////////////////

#define armEmitBranch(cond, link, address) \
	((cond) << 28) | (0x5 << 25) | ((link) << 24) | (address)

#define armEmitBranchX(cond, reg) \
	((cond) << 28) | (0x12 << 20) | (0xFFF1 << 4) | (reg)

/////////////////////////////////////////////////////////
// Multiplication commands
/////////////////////////////////////////////////////////

// rd = (rm * rs) + rn
#define armEmitMultiplyAccumulate(cond, s, rd, rn, rs, rm) \
	((cond) << 28) | (0x1 << 21) | ((s) << 20) | ((rd) << 16) | ((rn) << 12) | ((rs) << 8) | \
	(0x9 << 4) | (rm)

// rd = (rm * rs)
#define armEmitMultiply(cond, s, rd, rs, rm) \
	((cond) << 28) | ((s) << 20) | ((rd) << 16) | ((rs) << 8) | (0x9 << 4) | (rm)

/////////////////////////////////////////////////////////
// Status register commands
/////////////////////////////////////////////////////////

// r == 1 then SPSR is moved, otherwise CPSR
#define armEmitPSRToReg(cond, r, rd) \
	((cond) << 28) | (0x1 << 24) | ((r) << 22) | (0xf << 16) | ((rd) << 12)

// the only field needed is flags field (f), which is bit 3
#define armEmitPSRFlags 0x8
#define armEmitRegToPSRImm(cond, r, fields, rot, imm) \
	((cond) << 28) | (0x3 << 24) | ((r) << 22) | (0x1 << 21) | ((fields) << 16) | (0xf << 12) | \
	((rot) << 8) | (imm)

#define armEmitRegToPSRReg(cond, r, fields, rm) \
	((cond) << 28) | (0x1 << 24) | ((r) << 22) | (0x1 << 21) | ((fields) << 16) | (0xf << 12) | \
	(rm)

/////////////////////////////////////////////////////////
// Load/store multiple register commands
/////////////////////////////////////////////////////////

#define armEmitLoadStoreMultiple(cond, p, u, w, l, rn, regs) \
	((cond) << 28) | (0x1 << 27) | ((p) << 24) | ((u) << 23) | ((w) << 21) | ((l) << 20) | \
	((rn) << 16) | (regs)

/////////////////////////////////////////////////////////
// Software interrupt commands
/////////////////////////////////////////////////////////

#define armEmitSoftwareInterrupt(cond, immed) \
	((cond) << 28) | (0xF << 24) | (immed)

/////////////////////////////////////////////////////////
// Convienience definitions
/////////////////////////////////////////////////////////

#define ARM_R0 0
#define ARM_R1 1
#define ARM_R2 2
#define ARM_R3 3
#define ARM_R4 4
#define ARM_R5 5
#define ARM_R6 6
#define ARM_R7 7
#define ARM_R8 8
#define ARM_R9 9
#define ARM_R10 10
#define ARM_R11 11
#define ARM_R12 12
#define ARM_R13 13
#define ARM_LR 14
#define ARM_PC 15

#define ARM_FLAG_N 31
#define ARM_FLAG_Z 30
#define ARM_FLAG_C 29
#define ARM_FLAG_V 28
