#include "gbaspc.h"

u8 DSP_MEM[0x100] IN_EWRAM;
DspChannel channels[8] DATA_ALIGNED_EWRAM;

// Needs to be a power of two
#define NUM_SAMPLES 64
#define BRR_DECODEBUF_SIZE 0xAFFF

s16 decodedSampleBuffer[BRR_DECODEBUF_SIZE] IN_EWRAM;
u32 curSampleOffset IN_EWRAM;
u16 sampleLookup[NUM_SAMPLES] IN_EWRAM;
s16 *sampleAddress[NUM_SAMPLES] IN_EWRAM;
u16 sampleLengths[NUM_SAMPLES] IN_EWRAM;
u16 sampleLoopLengths[NUM_SAMPLES] IN_EWRAM;

#ifdef DSP_DEBUG
u32 dspBRRDecodes = 0;
#endif

void DecodeSample(DspChannel *channel, u16 sampStart, u16 sampLoop) {
	u16 lookupHash = sampStart & (NUM_SAMPLES -1);

	while (sampleLookup[lookupHash] != 0xffff &&
		sampleLookup[lookupHash] != sampStart) {
		lookupHash = (lookupHash + 7) & (NUM_SAMPLES -1);
	}

	if (sampleLookup[lookupHash] == sampStart) {
		channel->sample = sampleAddress[lookupHash];
		channel->sampleEnd = sampleAddress[lookupHash] + sampleLengths[lookupHash];
		if (sampleLoopLengths[lookupHash] != 0) {
			channel->loopStart = sampleAddress[lookupHash] + sampleLoopLengths[lookupHash];
			channel->flags = 1;
		} else {
			channel->flags = 0;
		}
		return;
	}

	s16 *sample = &(decodedSampleBuffer[curSampleOffset]);
	sampleLookup[lookupHash] = sampStart;
	sampleAddress[lookupHash] = sample;
	sampleLoopLengths[lookupHash] = 0;

#ifdef DSP_DEBUG
	dspBRRDecodes++;
#endif

	s8 *cur = &(APU_MEM[sampStart]);
	u16 length = 0;

	s16 prev1 = 0, prev2 = 0;
	do {
		s8 shift = 28 - ((*cur++ >> 4) & 0xF);

		if (shift == 28) { // if the range was 0
			for (int i=8; i!=0; i--) {
				*sample++ = (s16)(((s32)*cur >> 4) & -2);
				*sample++ = (s16)((((s32)*cur << 28) >> 28) & -2);
				cur++;
			}
		} else if (shift >= 16) { // if the range was 1 through 12
			for (int i=8; i!=0; i--) {
				*sample++ = (s16)((((s32)*cur & 0xF0) << 24) >> shift);
				*sample++ = (s16)(((s32)*cur << 28) >> shift);
				cur++;
			}
		} else { // invalid range
			for (int i=8; i!=0; i--) {
				*sample++ = (*cur & 0x80) ? -4096 : 0;
				*sample++ = (*cur & 8) ? -4096 : 0;
				cur++;
			}
		}

		sample -= 16;

		switch (cur[-9] & 0xC) {
		case 0x0:
			prev1 = sample[15];
			prev2 = sample[14];
			break;

		case 0x4:
			for (int i=16; i!=0; i--) {
				*sample += (s16)(prev1 + ((-prev1 >> 4) & -2));
				prev1 = *sample++;
			}
			prev2 = sample[-2];
			break;

		case 0x8: //Filter 2:  s = d + prev1(61/32) - prev2(15/16) 
			for (int i=16; i!=0; i--) 
			{ 
				s32 s = ((prev1 << 1) + (((-prev1 * 3) >> 5) & -2)) + 
					(-prev2 + ((prev2 >> 4) & -2)) + 
					*sample; 

				if (s < -65536) s = -65536;     //Clamp result to 17 bits 
				else if (s > 65534) s = 65534; 

				prev2 = prev1; 
				prev1 = *sample++ = (s16)s; 
			} 
			break; 

		case 0xC: //Filter 3:  s = d + prev1(115/64) - prev2(13/16) 
			for (int i=16; i!=0; i--) 
			{ 
				s32 s = ((prev1 << 1) + (((-13 * prev1) >> 6) & -2)) + 
					(-prev2 + (((prev2 * 3) >> 4) & -2)) + 
					*sample; 

				if (s < -65536) s = -65536; 
				else if (s > 65534) s = 65534; 

				prev2 = prev1; 
				prev1 = *sample++ = (s16)s; 
			} 
			break; 
		}

		length+=16;
	} while (!(cur[-9] & 1));

	if (cur[-9] & 2) {
		// Looping sample
		s32 loopLength = ((sampLoop - sampStart)/9) * 16;
		if (loopLength > 0 && loopLength <= length) {
			channel->flags = 1;
			channel->loopStart = sampleAddress[lookupHash] + loopLength;
			sampleLoopLengths[lookupHash] = loopLength;
		} else {
			channel->flags = 0;
		}
	} else {
		channel->flags = 0;
	}

	curSampleOffset += length;
	if (curSampleOffset > BRR_DECODEBUF_SIZE) {
		DrawString(0,0,"Too much decoded BRR data");
		for (;;);
	}

	sampleLengths[lookupHash] = length;

	channel->sample = sampleAddress[lookupHash];
	channel->sampleEnd = sampleAddress[lookupHash] + sampleLengths[lookupHash];
}

void DspReset() {
	curSampleOffset = 0;
	for (int i = 0; i < NUM_SAMPLES; i++) sampleLookup[i] = 0xffff;
	for (int i = 0; i < 8; i++) {
		channels[i].active = 0;
	}
}

#include "sprintf.h"
void dspunimpl(u8 opcode) {
	char c[50];
	sprintf(c,"DSP 0x%.2x broken",opcode);
	DrawString(0,0,c);
	for(;;);
}

void DspSetChannelVolume(u32 channel) {
	s8 leftVol = (DSP_MEM[(channel << 4) + DSP_VOL_L] * DSP_MEM[(channel << 4) + DSP_ENVX]) >> 7;
	s8 rightVol = (DSP_MEM[(channel << 4) + DSP_VOL_R] * DSP_MEM[(channel << 4) + DSP_ENVX]) >> 7;
	if (leftVol < 0) leftVol = -leftVol;
	if (rightVol < 0) rightVol = -rightVol;
	channels[channel].volume = (leftVol + rightVol) / 2;
}

void DspSetChannelPitch(u32 channel) {

	u16 rawFreq = *((u16*)&(DSP_MEM[(channel << 4) + DSP_PITCH_L]));

	u32 freq = (32000 * rawFreq) >> 12;

	#define MIXRATE 31536
	channels[channel].highSpeed = freq / MIXRATE;
	channels[channel].lowSpeed = (((freq - (MIXRATE * channels[channel].highSpeed)) << 16) / MIXRATE) << 16;
	channels[channel].count = 0;
}

void DspSetChannelSource(u32 channel) {
	u8 *sampleBase = APU_MEM + (DSP_MEM[DSP_DIR] << 8);
	u16 *lookupPointer = (u16*)(sampleBase + (DSP_MEM[(channel << 4) + DSP_SRC] << 2));
	DecodeSample(&(channels[channel]), lookupPointer[0], lookupPointer[1]);
}

void DspSetChannelEnvelope(u32 channel) {
	u8 adsr1 = DSP_MEM[(channel << 4) + DSP_ADSR1];
	u8 adsr2 = DSP_MEM[(channel << 4) + DSP_ADSR2];
	u8 gain = DSP_MEM[(channel << 4) + DSP_GAIN];

	if (adsr1 & 0x80) {
		// ADSR mode
		DSP_MEM[(channel << 4) + DSP_ENVX] = 0x7f;
	} else {
		// Gain mode
		if ((gain & 0x80) == 0) {
			// Direct designation
			DSP_MEM[(channel << 4) + DSP_ENVX] = gain & 0x7f;
			DspSetChannelVolume(channel);
		} else {
			DSP_MEM[(channel << 4) + DSP_ENVX] = 0x7f;
		}
	}
}

void DspWriteByte(u8 val, u8 address) {
	if (address > 0x7f) return;

	// Reads are mirrored, so duplicate the byte
	DSP_MEM[address] = val;
	DSP_MEM[0x80 | address] = val;

	switch (address) {
		case 0x00 + DSP_VOL_L:
		case 0x10 + DSP_VOL_L:
		case 0x20 + DSP_VOL_L:
		case 0x30 + DSP_VOL_L:
		case 0x40 + DSP_VOL_L:
		case 0x50 + DSP_VOL_L:
		case 0x60 + DSP_VOL_L:
		case 0x70 + DSP_VOL_L:
			DspSetChannelVolume(address >> 4);
			break;

		case 0x00 + DSP_VOL_R:
		case 0x10 + DSP_VOL_R:
		case 0x20 + DSP_VOL_R:
		case 0x30 + DSP_VOL_R:
		case 0x40 + DSP_VOL_R:
		case 0x50 + DSP_VOL_R:
		case 0x60 + DSP_VOL_R:
		case 0x70 + DSP_VOL_R:
			DspSetChannelVolume(address >> 4);
			break;

		case 0x00 + DSP_PITCH_L:
		case 0x10 + DSP_PITCH_L:
		case 0x20 + DSP_PITCH_L:
		case 0x30 + DSP_PITCH_L:
		case 0x40 + DSP_PITCH_L:
		case 0x50 + DSP_PITCH_L:
		case 0x60 + DSP_PITCH_L:
		case 0x70 + DSP_PITCH_L:
			DspSetChannelPitch(address >> 4);
			break;

		case 0x00 + DSP_PITCH_H:
		case 0x10 + DSP_PITCH_H:
		case 0x20 + DSP_PITCH_H:
		case 0x30 + DSP_PITCH_H:
		case 0x40 + DSP_PITCH_H:
		case 0x50 + DSP_PITCH_H:
		case 0x60 + DSP_PITCH_H:
		case 0x70 + DSP_PITCH_H:
			DspSetChannelPitch(address >> 4);
			break;

		case 0x00 + DSP_SRC:
		case 0x10 + DSP_SRC:
		case 0x20 + DSP_SRC:
		case 0x30 + DSP_SRC:
		case 0x40 + DSP_SRC:
		case 0x50 + DSP_SRC:
		case 0x60 + DSP_SRC:
		case 0x70 + DSP_SRC:
			// HACK: don't modify source while channel isn't playing, cause we don't want to decode
			// bogus samples for no reason
//			DspSetChannelSource(address >> 4);
			break;

		case 0x00 + DSP_ADSR1:
		case 0x10 + DSP_ADSR1:
		case 0x20 + DSP_ADSR1:
		case 0x30 + DSP_ADSR1:
		case 0x40 + DSP_ADSR1:
		case 0x50 + DSP_ADSR1:
		case 0x60 + DSP_ADSR1:
		case 0x70 + DSP_ADSR1:
			DspSetChannelEnvelope(address >> 4);
			break;

		case 0x00 + DSP_ADSR2:
		case 0x10 + DSP_ADSR2:
		case 0x20 + DSP_ADSR2:
		case 0x30 + DSP_ADSR2:
		case 0x40 + DSP_ADSR2:
		case 0x50 + DSP_ADSR2:
		case 0x60 + DSP_ADSR2:
		case 0x70 + DSP_ADSR2:
			DspSetChannelEnvelope(address >> 4);
			break;

		case 0x00 + DSP_GAIN:
		case 0x10 + DSP_GAIN:
		case 0x20 + DSP_GAIN:
		case 0x30 + DSP_GAIN:
		case 0x40 + DSP_GAIN:
		case 0x50 + DSP_GAIN:
		case 0x60 + DSP_GAIN:
		case 0x70 + DSP_GAIN:
			DspSetChannelEnvelope(address >> 4);
			break;

		case DSP_KON:
			for (int i=0; i<8; i++)
				if ((val>>i)&1) {
					DspSetChannelSource(i);
					DspSetChannelEnvelope(i);
					DspSetChannelVolume(i);
					DspSetChannelPitch(i);
			
					channels[i].active = 1;
					DSP_MEM[DSP_ENDX] &= ~(1 << i);
				}
			break;

		case DSP_KOF:
			for (int i=0; i<8; i++)
				if ((val>>i)&1) {
					DSP_MEM[DSP_KON] &= ~(1 << i);
					channels[i].active = 0;
				}
			break;

		case DSP_PMOD:
			for (int i=0; i<8; i++)
				if ((val>>i)&1) {
//					dspunimpl(DSP_PMOD);
//					channels[i].active = 0;
				}
			break;

		case DSP_NOV:
			for (int i=0; i<8; i++)
				if ((val>>i)&1) {
//					dspunimpl(DSP_NOV);
					channels[i].active = 0;
				}
			break;

		case DSP_ENDX:
			DSP_MEM[DSP_ENDX] = 0;
			DSP_MEM[0x80|DSP_ENDX] = 0;
			break;

		case DSP_DIR:
			// Once again, don't set source when we don't need to yet
/*			for (int i = 0; i < 8; i++) {
				DspSetChannelSource(i);
			}*/
			break;

		// Echo filter not supported
		case 0x0F:
		case 0x1F:
		case 0x2F:
		case 0x3F:
		case 0x4F:
		case 0x5F:
		case 0x6F:
		case 0x7F:
			break;

		default:
			break;
	}
}

void DspPrepareStateAfterReload() {
	for (u32 i = 0; i < 8; i++) {
		if (DSP_MEM[DSP_KON] & (1 << i)) {
			DspSetChannelSource(i);
			DspSetChannelPitch(i);
			DspSetChannelVolume(i);
			DspSetChannelEnvelope(i);

			channels[i].active = 1;
		}
		if (DSP_MEM[DSP_NOV] & (1 << i)) {
			channels[i].active = 0;
		}
	}
}

void DspSetupMixer() {
	for (u32 i = 0; i < 8; i++) {
		DspSetChannelVolume(i);
	}
}
